"""
    EghParams.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy._MOLASS.SerialSettings import get_setting
# from SecTheory.RetensionTime import make_initial_guess
from molass_legacy.SecTheory.SecEstimator import guess_initial_secparams, NUM_SEC_PARAMS
from .SimpleSecParams import SimpleSecParams, sec_comformance
from .BaselineParams import get_num_baseparams
from molass_legacy.Optimizer.BasicOptimizer import AVOID_VANISHING_RATIO

def get_common_parameter_names(n, nc_rg=None, seccol=True):
    if nc_rg is None:
        nc_rg = n
    else:
        # used in RtEmgParams
        pass

    nc = n - 1

    xr_names = []
    for k in range(nc):
        xr_names += ["$h_%d$" % k, r"$\mu_%d$" % k, r"$\sigma_%d$" % k, r"$\tau_%d$" % k]

    rg_names = ["$R_{g%d}$" % k for k in range(nc_rg)]

    mapping_names = ["$mp_a$", "$mp_b$"]

    uv_names = ["$uh_%d$" % k for k in range(nc)]

    mr_names = ["$mr_a$", "$mr_b$"]

    if seccol:
        seccol_names = ["$N_{pc}$", "$r_p$", "$t_I$", "$t_0$", "$P$", "$m$"]
    else:
        seccol_names = []

    return xr_names, rg_names, mapping_names, uv_names, mr_names, seccol_names

class EghParamsBase:
    def __init__(self, n_components, poresize, poreexponent, sec_class=None):
        self.n_components = n_components
        self.num_baseparams = get_num_baseparams()
        self.integral_baseline = self.num_baseparams == 3
        self.poresize = poresize
        if poreexponent == 0:
            # workaround for set_setting("poreexponent", None) not working
            poreexponent = None
        self.poreexponent = poreexponent
        if sec_class is None:
            sectype = SimpleSecParams(poresize, poreexponent)
        else:
            sectype = sec_class(poresize, poreexponent)
        self.init_method = sectype.init_method
        # self.conf_method = sectype.conf_method
        self.estm_method = sectype.estm_method
        self.nump_adjust = sectype.nump_adjust
        self.use_K = False

    def __str__(self):
        name = "Egh%sParams" % ("Advansed" if self.advanced else "")
        ps, px = [str(v) if v is None else "%g" % v for v in [self.poresize, self.poreexponent]]
        return "%s(nc=%d, ps=%s, px=%s)" % (name, self.n_components, ps, px)

    def get_model_name(self):
        return 'EGH'

    def compute_init_guess(self, *args):
        return self.init_method(*args)

    def compute_comformance(self, *args, **kwargs):
        # return self.conf_method(*args, **kwargs)       # bug in Python 3.10.7?
        return sec_comformance(*args, **kwargs)          # bug fix

    def get_seccol_params_for_disp(self, seccol_params):
        # recosider whether this is required
        return seccol_params

    def estimate_conformance_params(self, *args):
        return self.estm_method(*args)

    def get_xr_param_bounds(self, temp_xr_params, Ti, Np):      # Ti, Np not used
        xr_h_max = np.max(temp_xr_params[:,0])
        xr_h_min = xr_h_max*AVOID_VANISHING_RATIO
        m_allow = np.max(temp_xr_params[:,1])*0.2
        s_allow = np.average(np.abs(temp_xr_params[:,2]))*1.0
        xr_bounds = []
        for h, m, s, t in temp_xr_params:
            s = abs(s)
            xr_bounds.append((xr_h_min, xr_h_max*2))                # h
            xr_bounds.append((max(0, m - m_allow), m + m_allow))    # m
            xr_bounds.append((max(3, s - s_allow), s + s_allow))    # s
            xr_bounds.append((0, s + s_allow))                      # t
        return xr_bounds

    def get_extended_bounds(self, init_sec_params):
        Npc, rp, tI, t0, P, m = init_sec_params
        t0_upper_bound = get_setting("t0_upper_bound")
        if t0_upper_bound is None:
            t0_upper_bound = t0+500

        bounds = [(Npc-1, Npc+1), (rp-1, rp+1), (tI-500, tI+200), (t0-500, t0_upper_bound), (1000, 10000), (1, 3)]
        return bounds

    def get_param_bounds(self, params):
        split_params = self.split_params_simple(params, convert=False)      # note that init_xr_params will not be converted
        init_xr_params, init_xr_baseparams, init_rgs, init_mapping, init_uv_params, init_uv_baseparams, init_mappable_range, init_sec_params = split_params
        Npc, tI = init_sec_params[[0,2]]

        if self.advanced:
            temp_xr_params = init_xr_params
        else:
            temp_xr_params = np.vstack([init_xr_params, init_xr_baseparams])
        m_allow = np.max(temp_xr_params[:,1])*0.2

        xr_bounds = self.get_xr_param_bounds(temp_xr_params, tI, Npc)   # note that temp_xr_params is not converted

        if self.advanced:
            for k, v in enumerate(init_xr_baseparams):
                v_allow = max(0.1, abs(v))*0.2          # note that v may be zero
                if self.integral_baseline and k == 2:
                    bounds = (max(0, v - v_allow), v + v_allow)
                else:
                    bounds = (v - v_allow, v + v_allow)
                xr_bounds.append(bounds)

        rg_bounds = [(rg*0.5, rg*2) for rg in init_rgs]
        a, b = init_mapping
        mapping_bounds = [(a*0.8, a*1.2), (-m_allow, m_allow)]
        if self.advanced:
            temp_uv_params = init_uv_params
        else:
            temp_uv_params = np.concatenate([init_uv_params, init_uv_baseparams])
        uv_h_max = np.max(temp_uv_params)
        uv_h_min = uv_h_max*AVOID_VANISHING_RATIO
        uv_bounds = [(uv_h_min, uv_h_max*2) for h in temp_uv_params]
        if self.advanced:
            for k, v in enumerate(init_uv_baseparams):
                v_allow = max(0.1, abs(v))*0.2          # note that v may be zero
                if self.integral_baseline and k == 7:
                    bounds = (max(0, v - v_allow), v + v_allow)
                else:
                    bounds = (v - v_allow, v + v_allow)
                uv_bounds.append(bounds)

        f, t = init_mappable_range
        dx = (t - f)*0.1
        range_bounds = [(f-dx, f+dx), (t-dx, t+dx)]

        extended_bounds = self.get_extended_bounds(init_sec_params)

        self.bounds_lengths = [len(b) for b in [xr_bounds, rg_bounds, mapping_bounds, uv_bounds, range_bounds, extended_bounds]]
        return xr_bounds + rg_bounds + mapping_bounds + uv_bounds + range_bounds + extended_bounds

    def get_sigma_bounds(self, x, debug=True):
        # task: is this really necessary?
        xr_params = self.split_params_simple(x)[0]
        sigma = xr_params[:,2]
        tau = xr_params[:,3]
        max_sigma = np.max(np.sqrt(sigma**2 + tau**2))
        min_sigma = max_sigma/4
        return min_sigma, max_sigma*1.5

    def split_bounds(self, bounds):
        separate_bounds = []
        offset = 0
        for k, length in enumerate(self.bounds_lengths):
            next_offset = offset + length
            sep_bounds = bounds[offset:next_offset]
            if k == 0:
                sep_bounds = self.reshape_xr_bounds(sep_bounds)
            separate_bounds.append(sep_bounds)
            offset = next_offset
        return separate_bounds

    def update_bounds_hook(self, masked_init_params):
        for i in range(self.n_components - 1):
            j = 3*i + 2     # sigma index
            masked_init_params[j] = max(1, masked_init_params[j])   # sigma should not be less than 1

    def get_peak_pos_array_list(self, x_array):
        n = self.n_components - 1
        pos_array_list = []
        for k in range(n):
            pos_array_list.append(x_array[:,4*k+1])     # mu in (h, mu, sigma, tau)
        return pos_array_list

    def get_params_sheet(self, parent, params, dsets, optimizer, debug=True):
        if debug:
            from importlib import reload
            import molass_legacy.ModelParams.EghParamsSheet
            reload(molass_legacy.ModelParams.EghParamsSheet)
        from .EghParamsSheet import EghParamsSheet
        return EghParamsSheet(parent, params, dsets, optimizer)

    def get_adjuster(self, debug=True):
        if debug:
            import molass_legacy.ModelParams.EghAdjuster
            from importlib import reload
            reload(molass_legacy.ModelParams.EghAdjuster)
        from .EghAdjuster import EghAdjuster
        return EghAdjuster()

    def compute_xr_peak_position(self, i, xr_params):
        return xr_params[i,1]

class EghParams(EghParamsBase):
    def __init__(self, n_components, poresize=None, poreexponent=None, sec_class=None):
        EghParamsBase.__init__(self, n_components, poresize, poreexponent, sec_class=sec_class)
        self.num_params = 6*n_components + 4
        self.advanced = False

    def split_params(self, n, p):
        if self.num_params != len(p):
            raise ValueError("len(params)=%d != %d, which is calculated from number of components %d." % (len(p), self.num_params, n))

        sep = 4*(n-1)
        xr_params = p[0:sep].reshape((n-1,4))
        xr_baseparams = p[sep:sep+4]
        sep += 4
        rgs = p[sep:sep+n]
        sep += n
        mapping = p[sep:sep+2]
        sep = sep+2
        uv_params = p[sep:sep+n-1]
        sep += n-1
        uv_baseparams = p[sep:sep+1]    # i.e., size one vector
        sep += 1
        mappable_range = p[sep:]

        return xr_params, xr_baseparams, rgs, mapping, uv_params, uv_baseparams, mappable_range

    def split_params_simple(self, params, convert=False):       # convert is not relevant in free EGH
        n = self.n_components
        r = len(params) - self.num_params
        if r > 0:
            decomp_params = params[:-r]
            seccol_params = params[-r:]
        else:
            decomp_params = params
            seccol_params = None
        self.separate_params = self.split_params(n, decomp_params) + [seccol_params]
        return self.separate_params

    def make_bounds_mask(self):
        bounds_mask = np.zeros(self.num_params, dtype=bool)
        n = self.n_components
        xr_baseparams_mask = [False]*4      # no bounds for baseline params
        bounds_mask[0:n*5] = np.array([True, True, True, False]*(n-1) + xr_baseparams_mask
            + [True]*(n-1) + [False])       # no bounds for the last rg of baseline
        return bounds_mask

    def get_parameter_names(self):
        xr_names, rg_names, mapping_names, uv_names, mr_names, seccol_names = get_common_parameter_names(self.n_components)
        xr_basenames = ["$xb_h$", "$xb_a$", "$xb_b$", "$\tau_u$"]
        uv_basenames = ["$uh_b$"]
        return np.array(xr_names + xr_basenames + rg_names + mapping_names + uv_names + uv_basenames + mr_names + seccol_names)

    def get_rg_start_index(self):
        return 4*self.n_components

    def get_mr_start_index(self):
        return 6*self.n_components + 2

class EghAdvansedParams(EghParamsBase):
    def __init__(self, n_components, poresize=None, poreexponent=None, sec_class=None, baseline_rg=True):
        EghParamsBase.__init__(self, n_components, poresize, poreexponent, sec_class=sec_class)
        self.advanced = True
        self.baseline_rg = baseline_rg

        nc = n_components - 1

        self.pos = []
        self.pos.append(0)      # [0] xr_params
        sep = 4*nc
        self.pos.append(sep)    # [1] xr_baseparams
        sep += self.num_baseparams
        self.pos.append(sep)    # [2] rgs
        if self.baseline_rg:
            # for backward compatibility toward _MOLASS 2.0.0*
            sep += nc + 1                   # note that, due to historical reasons, len(rgs) = n_components,
                                            # i.e., the extra rg from molass_legacy.Baseline is included, which is not used
        else:
            # for _MOLASS 2.1.0* or later
            sep += nc
        self.pos.append(sep)    # [3] mapping
        sep = sep+2
        self.pos.append(sep)    # [4] uv_params
        sep += nc
        self.pos.append(sep)    # [5] uv_baseparams
        sep += 5 + self.num_baseparams
        self.pos.append(sep)    # [6] mappable_range
        sep += 2
        self.pos.append(sep)    # [7] end

        self.num_params = sep   # note that this does not include sec params

    def get_estimator(self, editor, debug=True):
        if debug:
            from importlib import reload
            import molass_legacy.Estimators.EghEstimator
            reload(molass_legacy.Estimators.EghEstimator)
        from molass_legacy.Estimators.EghEstimator import EghEstimator
        return EghEstimator(editor)

    def split_params(self, n, params):
        if self.num_params != len(params):
            raise ValueError("len(params)=%d != %d, which is calculated from number of components %d." % (len(params), self.num_params, n))

        ret_params = []
        for p, q in zip(self.pos[:-1], self.pos[1:]):
            ret_params.append(params[p:q])

        ret_params[0] = ret_params[0].reshape((n-1,4))  # xr_params

        return ret_params

    def split_params_simple(self, params, convert=False):       # convert is not relevant in free EGH
        n = self.n_components
        r = len(params) - self.num_params
        if r > 0:
            decomp_params = params[:-r]
            seccol_params = params[-r:]
        else:
            decomp_params = params
            seccol_params = None
        return self.split_params(n, decomp_params) + [seccol_params]

    def split_as_unified_params(self, params, **kwargs):
        return self.split_params_simple(params, **kwargs) + [(None, None)]

    def make_bounds_mask(self):
        bounds_mask = np.zeros(self.num_params + NUM_SEC_PARAMS, dtype=bool)
        n = self.n_components
        xr_base_start = 4*(n-1)
        bounds_mask[0:xr_base_start] = np.array([True, True, True, False]*(n-1))

        rg_start = xr_base_start + self.num_baseparams

        if self.integral_baseline:
            bounds_mask[rg_start-1] = True      # xr baseline fouling
            bounds_mask[self.pos[6]-1] = True   # uv baseline fouling

        bounds_mask[rg_start:rg_start+n] = np.array([True]*(n-1) + [False])    # no bounds for the last rg of baseline
        return bounds_mask

    def get_parameter_names(self):
        xr_names, rg_names, mapping_names, uv_names, mr_names, seccol_names = get_common_parameter_names(self.n_components)
        xr_basenames = ["$xb_a$", "$xb_b$"]
        if self.num_baseparams == 3:
            xr_basenames += ["$xb_r$"]
        uv_basenames = ["$L$", "$x_0$", "$k$", "$b$", "$s_1$", "$s_2$", "$diffratio$"]
        if self.num_baseparams == 3:
            uv_basenames += ["$ub_r$"]
        return np.array(xr_names + xr_basenames + rg_names + mapping_names + uv_names + uv_basenames + mr_names + seccol_names)

    def get_rg_start_index(self):
        return self.pos[2]

    def get_mr_start_index(self):
        return self.pos[6]

    def split_get_unified_sec_params(self, params):
        t0, K, rp, m = self.split_params_simple(params)[-1]
        return (t0, K, rp, m) + (None,)*7

def construct_egh_params_type(n_components, sec_class=None, baseline_rg=True):
    from molass_legacy._MOLASS.SerialSettings import get_setting

    poresize = get_setting("poresize")
    poreexponent = get_setting("poreexponent")
    uv_basemodel = get_setting("uv_basemodel")
    if uv_basemodel == 1:
        params_type = EghAdvansedParams(n_components, poresize, poreexponent, sec_class=sec_class, baseline_rg=baseline_rg)
    else:
        params_type = EghParams(n_components, poresize, poreexponent, sec_class=sec_class)
    return params_type



"""
    TkMplUtils.py

    leanred from Cody AI on 2024-02-02
"""

def adjust_the_tkframe_size(fig, canvas_widget, tkframe):
    """
    Adjust the size of the tkinter frame to fit the figure.
    We need this because this synchronization is not yet supported by FigureCanvasTkAgg.
    """
    w, h = fig.canvas.get_width_height()
    print("adjust_the_tkframe_size: w, h =", w, h)
    tkframe.config(width=w, height=h)
    canvas_widget.config(width=w, height=h)


"""
    DynamicObjeccts.py

    Copyright (c) 2018, Masatsuyo Takahashi
"""

class FlexibleStruct(object):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __getattr__(self, attr):
        setattr(self, attr, FlexibleStruct())
        return getattr(self, attr)



# coding: utf-8
"""
    DualSpace.py

    Copyright (c) 2019,2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.patches import Polygon
import mpl_toolkits.mplot3d.art3d as art3d
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
import molass_legacy.KekLib.DebugPlot as dplt
from molass_legacy.KekLib.OurMatplotlib import get_color
from ModeledData import ModeledData
from MatrixData import simple_plot_3d
from OurManim import manim_init
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar

class DualSpace(Dialog):
    def __init__(self, parent):
        self.parent = parent
        self.currently_dragging = False
        self.current_n = None
        self.polygon = None
        manim_init()
        Dialog.__init__(self, parent, "Dual Space", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()
        bframe = Tk.Frame(body_frame)
        bframe.pack(fill=Tk.BOTH)
        tframe = Tk.Frame(bframe)
        tframe.pack(side=Tk.LEFT)

        fig = plt.figure(figsize=(21,7))
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

        ax1 = fig.add_subplot(131, projection='3d')
        ax2 = fig.add_subplot(132)
        ax3 = fig.add_subplot(133)
        self.axes = ax1, ax2, ax3
        self.draw()
        fig.tight_layout()

        self.mpl_canvas.draw()

    def draw(self):
        q = np.linspace(0.01, 0.6, 600)
        rg_list = [50, 24]
        d_list = [1, 3]
        md = ModeledData(q, 300, rg_list=rg_list, d_list=d_list)

        ax1, ax2, ax3 = self.axes

        ax1.set_xlabel('Q')
        ax1.set_ylabel('Eno')
        ax1.set_zlabel('Intensity')

        M = md.get_data()
        simple_plot_3d(ax1, M, x=q)
        ax1.view_init(elev=20, azim=-20)

        i = [20, 130]
        j = [120, 170]

        n = M.shape[1]
        ey = np.arange(n)
        for i_ in i:
            ex = np.ones(n)*q[i_]
            ez = M[i_,:]
            # ax1.plot(ex, ey, ez, color='orange')
            poly_points = [(0,0)] + list(zip(ey,ez)) + [(n-1,0)]
            poly =  Polygon(poly_points, alpha=0.3, color='orange')
            ax1.add_patch(poly)
            art3d.pathpatch_2d_to_3d(poly, z=q[i_], zdir='x')

        n = M.shape[0]
        ex = q
        for j_ in j:
            ey = np.ones(n)*j_
            ez = M[:,j_]
            ax1.plot(ex, ey, ez, color='green')
            poly_points = [(0,0)] + list(zip(ex,ez)) + [(q[-1],0)]
            poly =  Polygon(poly_points, alpha=0.3, color='green')
            ax1.add_patch(poly)
            art3d.pathpatch_2d_to_3d(poly, z=j_, zdir='y')

        ax2.set_title('2D Scattering Curve Space', fontsize=20)
        ax2.set_xlabel('Intensity on Q=%.2g' % q[i[0]], color='blue', fontsize=16)
        ax2.set_ylabel('Intensity on Q=%.2g' % q[i[1]], color='red', fontsize=16)

        labels = ['a', 'b', 'c', 'd']
        x = [q[i[0]], q[i[1]], q[i[0]], q[i[1]]]
        y = [j[0], j[0], j[1], j[1]]
        z = [M[i[0],j[0]], M[i[1],j[0]], M[i[0],j[1]], M[i[1],j[1]]]
        k = 0
        for label, x_, y_, z_ in zip(labels, x, y, z):
            color = 'blue' if k%2==0 else 'red'
            ax1.plot([x_, x_], [y_, y_], [0, z_], color=color, linewidth=3)
            ax1.text(x_, y_, z_, label, fontsize=20, color=color)
            k += 1

        a, b, c, d = z
        s_points = [(0,0), (1,0), (1,1), (0,1)]
        p_points = [(0,0), (a,b), (a+c, b+d), (c,d)]
        for points, color in [(s_points, 'pink'), (p_points, 'cyan')]:
            poly =  Polygon(points, alpha=0.3, color=color)
            ax2.add_patch(poly)

        def plot_point_s(x, y, label):
            ax2.plot([0,x], [y,y], color='blue')
            ax2.plot([x,x], [0,y], color='red')
            ax2.plot(x, y, 'o', color='orange')
            ax2.text(x, y+0.04, label, fontsize=20)

        plot_point_s(a, b, r"$ \begin{bmatrix} a \\ b \end{bmatrix} $")
        plot_point_s(c, d, r"$ \begin{bmatrix} c \\ d \end{bmatrix} $")
        ax2.set_xlim(-0.4, 1.4)
        ax2.set_ylim(-0.4, 1.4)

        ax3.set_title('2D Elution Curve Space', fontsize=20)
        ax3.set_xlabel('Intensity on Eno=%d' % j[0], color='purple', fontsize=16)
        ax3.set_ylabel('Intensity on Eno=%d' % j[1], color='purple', fontsize=16)

        s_points = [(0,0), (1,0), (1,1), (0,1)]
        p_points = [(0,0), (a,c), (a+b, c+d), (b,d)]
        for points, color in [(s_points, 'pink'), (p_points, 'cyan')]:
            poly =  Polygon(points, alpha=0.3, color=color)
            ax3.add_patch(poly)

        def plot_point_e(x, y, label, color):
            ax3.plot([0,x], [y,y], color=color)
            ax3.plot([x,x], [0,y], color=color)
            ax3.plot(x, y, 'o', color='orange')
            ax3.text(x, y+0.04, label, fontsize=20)

        plot_point_e(a, c, r"$ \begin{bmatrix} a \\ c \end{bmatrix} $", "blue")
        plot_point_e(b, d, r"$ \begin{bmatrix} b \\ d \end{bmatrix} $", "red")

        ax3.set_xlim(-0.4, 1.4)
        ax3.set_ylim(-0.4, 1.4)



# DPHEM library which is compatible with sklearn.mixture.GaussianMixture 
# ---
# If you use this code, please cite our paper:
#   Lei Yu, Tianyu Yang, and Antoni B. Chan. "Density Preserving Hierarchical EM Algorithm: Simplifying 
#   Gaussian Mixture Models for Approximate Inference." TPAMI 2019
# ---
# Copyright (c) 2019-04-30
# Lei Yu, Antoni B. Chan
# Department of Computer Science
# City University of Hong Kong 
#
# v0.1


import numpy as np
from sklearn.mixture import GaussianMixture
from .pylib_dphem import *
import warnings


def gmm_dphem_learn(basegmm, tagk, opt):
    # extract the GMM parameters to implement our algorithm

    if 'showplot' not in opt:
        opt['showplot'] = 0

    if 'LLterm' not in opt:
        opt['LLterm'] = 1e-5

    if 'cvminreg' not in opt:
        opt['cvminreg'] = 1e-6

    if 'verbose' not in opt:
        opt['verbose'] = 1

    if 'pimode' not in opt:
        opt['pimode'] = 'preserve'

    if 'maxiter' not in opt:
        opt['maxiter'] = 1000

    if 'Nvs' not in opt:
        opt['Nvs'] = 10*basegmm.n_components

    maxiter = opt['maxiter']
    showplot = opt['showplot']
    ncomp_base,dim = np.shape(basegmm.means_)  
                                    

    # Initialization of reducedgmm
    # 0. fixed initgmm with some given GMM
    # 1. randomly select tagk components from the basegmm
    # 2. weighted kmeans of base mixture components
    # 3. splitting from one component
    if opt['initmode'] == 'fix':
        if 'initgmm' not in opt or not isinstance(opt['initgmm'],GaussianMixture):
            raise ValueError('Init GMM with GaussianMixture type should be given.')
        else:
            gmm = opt['initgmm']

        # with an init reducedgmm, we'll iterate with EM to optimize the reducedgmm
        num_iter = 0
        dataLLall = []
        dataLL = 0.0

        while True:
            dataLLold = dataLL
            LL,LLcomp,Y,LB,entropyLB = E_step(basegmm,gmm,opt)

            dataLL = LB

            dataLLall.append(dataLL)

            do_break = False

            if num_iter > 0:
                dLL = dataLL - dataLLold
                pLL = np.fabs(dLL/dataLL)
                if opt['verbose']:
                    print(('iter = {0}; LL = {1}; dLL = {2}; pLL = {3} \n'.
                        format(num_iter, dataLL, dLL, pLL)))

                if dLL < 0:
                    print(('\nnum_iter = {0}\n'.format(num_iter)))
                    warnings.warn('LL change is negative.')

                if pLL < opt['LLterm']:
                    do_break = True

            if num_iter >= maxiter:
                warnings.warn('Max iterations reached.')
                do_break = True

            if do_break:
                break

            gmm = M_step(gmm, basegmm, Y, opt)
            num_iter += 1

        gpost = Y
        info = {'dataLL': dataLL,
                'dataLLall': np.asarray(dataLLall),
                'opt': opt}
        diffel = entropyLB - LB

        return gmm, gpost, info, diffel


    if opt['initmode'] == 'random':
        # randomly select tagk components from the basegmm as init reducedgmm
        # only means are the same but weights and covariances are not
        initgmm = GaussianMixture(n_components=tagk,covariance_type=basegmm.covariance_type)
        weights = np.divide(np.ones(tagk),tagk)
        means = []
        covariances = []

        cvall = 0
        for j in range(ncomp_base):
            if (basegmm.covariance_type == 'spherical' or 
               basegmm.covariance_type == 'diag'):
               cvall += np.mean(basegmm.covariances_[j])
            elif basegmm.covariance_type == 'full':
                cvall += np.mean(np.diagonal(basegmm.covariances_[j]))
            else:
                raise ValueError('Undefined covariance type.')

        cvall = float(cvall)/ncomp_base

        foo = np.random.permutation(list(range(ncomp_base)))

        for j in range(tagk):
            means.append(basegmm.means_[foo[j]])
            if basegmm.covariance_type == 'spherical':
                covariances.append(cvall)
            elif basegmm.covariance_type == 'diag':
                covariances.append(cvall*np.ones(dim))
            elif basegmm.covariance_type == 'full':
                covariances.append(cvall*np.eye(dim))
            else:
                raise ValueError('Undefined covariance type.')

        initgmm.weights_ = weights
        initgmm.means_   = np.asarray(means)
        initgmm.covariances_ = np.asarray(covariances)

        opt['initmode'] = 'fix'
        opt['initgmm']  = initgmm
        reducedgmm, gpost, info, diffel = gmm_dphem_learn(basegmm, tagk, opt)

        return reducedgmm, gpost, info, diffel


    if opt['initmode'] == 'wtkmeans':
        # Initialize the reducedgmm with weighted kmeans of the basegmm components
        initgmm = GaussianMixture(n_components=tagk,covariance_type=basegmm.covariance_type)
        weights = np.divide(np.ones(tagk),tagk)
        covariances = []  # use the average covariances of basegmm components

        # randomly select tagk components as the cluster centers
        foo = np.random.permutation(list(range(ncomp_base)))
        init_cluster_center = []
        for i in range(tagk):
            init_cluster_center.append(basegmm.means_[foo[i]])

        init_cluster_center = np.asarray(init_cluster_center).T
        cluster_label,cluster_center = my_weighted_kmeans(basegmm.means_.T,
             basegmm.weights_, init_cluster_center, tagk, 100)
        means = cluster_center.T

        cvall = 0
        for j in range(ncomp_base):
            if (basegmm.covariance_type == 'spherical' or 
               basegmm.covariance_type == 'diag'):
               cvall += np.mean(basegmm.covariances_[j])
            elif basegmm.covariance_type == 'full':
                cvall += np.mean(np.diagonal(basegmm.covariances_[j]))
            else:
                raise ValueError('Undefined covariance type.')

        cvall = float(cvall)/ncomp_base

        for j in range(tagk):
            if basegmm.covariance_type == 'spherical':
                covariances.append(cvall)
            elif basegmm.covariance_type == 'diag':
                covariances.append(cvall*np.ones(dim))
            elif basegmm.covariance_type == 'full':
                covariances.append(cvall*np.eye(dim))
            else:
                raise ValueError('Undefined covariance type.')

        initgmm.weights_ = weights
        initgmm.means_   = means
        initgmm.covariances_ = covariances

        opt['initmode'] = 'fix'
        opt['initgmm']  = initgmm
        reducedgmm,gpost,info,diffel = gmm_dphem_learn(basegmm, tagk, opt)

        return reducedgmm,gpost,info,diffel


    


